// [UXLĩV~[VvO@by Kosei Demura (2010-7-31)
// @
// fL[F@OiC@bL[FiCrL[FE]C@lL[F], sL[F~
// vL[F@[U̕\ON/OFF؂ւ

#include <ode/ode.h>
#include <drawstuff/drawstuff.h>
#include "main.h"
#include <sys/time.h>

/*** [U[̔ˁ@***/
static void emitLaser()
{
    dMatrix3 R1, R2, R3, R4, R5;

    const dReal *p = (const dReal *) dBodyGetPosition(laser.body);
    dReal heading_angle = -heading();
    dReal pitch  = dJointGetHingeAngle(laser_joint);

    dRFromAxisAndAngle(R1,1,0,0, -M_PI/2);
    dRFromAxisAndAngle(R4,1,0,0, pitch);

    for (int i=urg.first_index; i < urg.last_index; i+=urg.skip_lines)
    {
        for (int j=0; j < 3; j++)
        {
            contact_pos[i][j] = -1;
        }

        dRFromAxisAndAngle(R2,0,0,1,
                           heading_angle +(double) (i - (BEAM_NUM/2))/(BEAM_NUM/2) * LASER_LO_LIMIT * (M_PI/180));
        dMultiply0(R3,R2,R1,3,3,3);
        dMultiply0(R5,R4,R3,3,3,3);
        dGeomSetPosition(beam[i], p[0], p[1], p[2]);
        dGeomSetRotation(beam[i],R5);
    }
}

/*** [UXLi̍쐬 ***/
static void makeLaserScanner()
{
    dMass mass;

    // [UXLi@
    laser.body  = dBodyCreate(world);
    dMassSetZero(&mass);
    dMassSetCylinderTotal(&mass,LASER_M,3,LASER_R, LASER_L);
    dBodySetMass(laser.body,&mass);
    dBodySetPosition(laser.body,START_X, START_Y, START_Z+NECK_L+SENSOR_L+LASER_L/2);

    // [Ur[DRayWIgŎD{fB͂ȂD
    for (int i=0; i < BEAM_NUM; i++)
    {
        beam[i] =  dCreateRay(space, 0);
        dGeomRaySetLength(beam[i], BEAM_LENGTH);
    }
}


static void makeRobot()
{
    dMatrix3 R;
    dMass mass, mass2, mass3;

    // 
    base.body  = dBodyCreate(world);
    dMassSetZero(&mass);
    dMassSetCylinderTotal(&mass,BASE_M,3,BASE_R,NECK_L);
    dBodySetMass(base.body,&mass);
    base.geom = dCreateCylinder(space,BASE_R,BASE_L);
    dGeomSetBody(base.geom,base.body);
    dBodySetPosition(base.body,START_X,START_Y,START_Z);

    // 
    neck.body  = dBodyCreate(world);
    dMassSetZero(&mass2);
    dMassSetCylinderTotal(&mass2,NECK_M,3,NECK_R,NECK_L);
    dBodySetMass(neck.body,&mass2);

    neck.geom = dCreateCylinder(space,NECK_R,NECK_L);
    dGeomSetBody(neck.geom,neck.body);
    dBodySetPosition(neck.body,START_X, START_Y, START_Z+0.5*NECK_L);
    dGeomDisable(neck.geom);

    // ZT
    sensor.body  = dBodyCreate(world);
    dMassSetZero(&mass3);
    dMassSetCylinderTotal(&mass3,SENSOR_M,3,SENSOR_R, SENSOR_L);
    dBodySetMass(sensor.body,&mass3);

    sensor.geom = dCreateCylinder(space,SENSOR_R, SENSOR_L);
    dGeomSetBody(sensor.geom,sensor.body);
    dBodySetPosition(sensor.body,START_X, START_Y, START_Z+NECK_L);
    dGeomDisable(sensor.geom);

    // [UXLi̍쐬
    makeLaserScanner();

    // ԗ
    dRFromAxisAndAngle(R,0,1,0,M_PI/2.0);
    for (int i = 0; i < NUM; i++)
    {
        wheel[i].body = dBodyCreate(world);

        dMassSetZero(&mass);
        if ((i % 2) == 0)
        {
            dMassSetCylinderTotal(&mass,WHEEL_SMALL_M, 1, 0.95*WHEEL_SMALL_R, WHEEL_W);
            dBodySetMass(wheel[i].body,&mass);
            wheel[i].geom = dCreateCylinder(space, 0.95*WHEEL_SMALL_R, WHEEL_W);
        }
        else
        {
            dMassSetCylinderTotal(&mass,WHEEL_BIG_M, 1, WHEEL_BIG_R, WHEEL_W);
            dBodySetMass(wheel[i].body,&mass);
            wheel[i].geom = dCreateCylinder(space, WHEEL_BIG_R, WHEEL_W);
        }
        dGeomSetBody(wheel[i].geom, wheel[i].body);
        dBodySetRotation(wheel[i].body,R);
    }

    dReal w1x = 0.5 * (BASE_R * 2.0 + WHEEL_W);
    dReal w1z = START_Z;

    dReal w0y = 0.5 * (BASE_R * 2.0) - WHEEL_SMALL_R;
    dReal w0z = START_Z - 0.5 * BASE_L - WHEEL_SMALL_R;

    dBodySetPosition(wheel[FRONT].body,  START_X,     START_Y+w0y, w0z);
    dBodySetPosition(wheel[LEFT].body,   START_X-w1x, START_Y,     w1z);
    dBodySetPosition(wheel[REAR].body,   START_X,     START_Y-w0y, w0z);
    dBodySetPosition(wheel[RIGHT].body,  START_X+w1x, START_Y,     w1z);

    // qWWCg Ԏp
    for (int i = 0; i < NUM; i++)
    {
        joint[i] = dJointCreateHinge(world,0);
        dJointAttach(joint[i], base.body, wheel[i].body);
    }
    dJointSetHingeAxis(joint[FRONT],1, 0, 0);
    dJointSetHingeAxis(joint[LEFT], 1, 0, 0);
    dJointSetHingeAxis(joint[REAR], 1, 0, 0);
    dJointSetHingeAxis(joint[RIGHT],1, 0, 0);
    dJointSetHingeAnchor(joint[FRONT], START_X,     START_Y+w0y, w0z); // front
    dJointSetHingeAnchor(joint[LEFT],  START_X-w1x, START_Y,     w1z); // left
    dJointSetHingeAnchor(joint[REAR],  START_X,     START_Y-w0y, w0z); // rear
    dJointSetHingeAnchor(joint[RIGHT], START_X+w1x, START_Y,     w1z); // right

    // ZTxpŒWCg
    neck_joint = dJointCreateFixed(world,0);
    dJointAttach(neck_joint,neck.body,base.body);
    dJointSetFixed(neck_joint);

    // ZTŒpWCg
    sensor_joint = dJointCreateFixed(world,0);
    dJointAttach(sensor_joint,neck.body,sensor.body);
    dJointSetFixed(sensor_joint);

    // qWWCg@[UXLip
    laser_joint = dJointCreateHinge(world,0);
    dJointAttach(laser_joint,sensor.body,laser.body);
    dJointSetHingeAxis(laser_joint,-1,0,0);
    //dJointSetHingeAnchor(sensor_joint,START_X, START_Y+0.5*BASE_S[1]-0.1*SENSOR_S[1],START_Z+NECK_L);
    dJointSetHingeAnchor(laser_joint,START_X, START_Y,START_Z+NECK_L+SENSOR_L);
    dJointSetHingeParam(laser_joint,dParamLoStop, -M_PI/4);
    dJointSetHingeParam(laser_joint,dParamHiStop,  M_PI/4);
}


/*** S[̐ ***/
static void makeWall()
{
    dReal pos[WALL_PARTS_NUM][3] = {{1.5, 1.5, 1.0},{ -1.5, 1.5, 1.0},
        {1.5, 5.5, 1.0},{-1.5, 5.5, 1.0}
    };

    for (int i = 0; i < WALL_PARTS_NUM; i++)
    {
        wall_parts[i] = dCreateBox(space, WALL_SIDES[i][0],
                                   WALL_SIDES[i][1], WALL_SIDES[i][2]);
        dGeomSetPosition(wall_parts[i],pos[i][0], pos[i][1], pos[i][2]);
    }


}

/*** [Ũf[^擾@***/
void laser_receiveData(laser_t *urg, long data[], int data_max)
{
    const dReal *pos;

    pos = dBodyGetPosition(laser.body);
    for (int i = 0; i < BEAM_NUM; i++)
    {
        if (!((contact_pos[i][0] == -1) && (contact_pos[i][1] == -1) &&
                (contact_pos[i][2] == -1)))
        {
            // `ɂ邽߂ɊԈ
            dReal d2 = 0;
            for (int j = 0; j < 3; j++)
            {
                d2 += (pos[j]-contact_pos[i][j]) * (pos[j] - contact_pos[i][j]);
            }
            data[i] = (long) (1000 * sqrt(d2));
            // printf("Laser dist[%d]=%.2f \n",i,laser_dist[i]);
        }
        else
        {
            data[i] = -1;
        }
    }
}

void laser_setSkipLines(laser_t *urg, int skip_lines)
{
    urg->skip_lines = skip_lines;
}

void laser_requestData(laser_t *urg, laser_request_type request_type, int first_index, int last_index)
{
    if ((first_index < 0)||(first_index > BEAM_NUM)|| (first_index > last_index))
    {
        printf("first index error \n");
    }
    else
    {
        urg->first_index = first_index;
    }

    if ((last_index < 0) || (last_index > BEAM_NUM) || (last_index < first_index))
    {
        printf("last index error \n");
    }
    else
    {
        urg->last_index  = last_index;
    }
}


/*** [U̕`@***/
void drawLaser()
{
    const dReal *pos;
    dsSetColorAlpha(1.0, 0.0, 0.0, 0.5);
    static int skip = 6; // `ɂ邽߂̊Ԉ

    pos = dBodyGetPosition(laser.body);
    for (int i = 0; i < BEAM_NUM; i++)
    {
        if (!((contact_pos[i][0] == -1) && (contact_pos[i][1] == -1) &&
                (contact_pos[i][2] == -1)))
        {
            // `ɂ邽߂ɊԈ
            if ((i % skip) == 0)dsDrawLine(pos,(const double *) contact_pos[i]);
        }
    }
}

/*** {[̕` ***/
void drawBall()
{
    dsSetColor(1.0,0.0,0.0);
    dsDrawSphere((const dReal *) dGeomGetPosition(ball.geom),
                 dGeomGetRotation(ball.geom),dGeomSphereGetRadius(ball.geom));
}

/*** S[̕` ***/
static void drawWall()
{
    dsSetTexture(DS_NONE);
    for (int i = 0; i < WALL_PARTS_NUM; i++)
    {
        dsSetColor(1.3, 1.3, 1.3);

        dsDrawBox(dGeomGetPosition(wall_parts[i]),
                  dGeomGetRotation(wall_parts[i]),WALL_SIDES[i]);

    }
}

/*** ~̕` ***/
static void drawCircle(dReal r,dReal center[2], int angle1, int angle2)
{
    dReal pos1[3],pos2[3],z=0.005;
    pos1[0] = r * cos(angle1 * M_PI/180.0) + center[0];
    pos1[1] = r * sin(angle1 * M_PI/180.0) + center[1];
    pos1[2] = z;

    for (int i = angle1; i < angle2; i++)
    {
        pos2[0] = r* cos(i * M_PI/180.0) + center[0];
        pos2[1] = r* sin(i * M_PI/180.0) + center[1];
        pos2[2] = z;
        dsDrawLine(pos1,pos2);
        pos1[0] = pos2[0];
        pos1[1] = pos2[1];
        pos1[2] = pos2[2];
    }
}


/*** ̕` ***/
static void drawCourse()
{
    dReal z = 0.005;
    dReal center_r = 1.0, corner_r = 0.5;
    dReal lx = 10, ly = 10;
    dReal pos[][3] = {{lx,ly,z},{-lx,ly,z},{-lx,-ly,z},{lx,-ly,z}};
    dsSetColor(1.3,1.3,1.3);
    dsDrawLine(pos[0],pos[1]);
    dsDrawLine(pos[1],pos[2]);
    dsDrawLine(pos[2],pos[3]);
    dsDrawLine(pos[3],pos[0]);

    // X^[gT[N̕`
    dReal pos_center[2] = {50,50};
    drawCircle(center_r,pos_center, 0, 360);

}

/*** X^[g֐ ***/
static void start()
{
    float xyz[3] = {   2.7f, -1.8f, 2.1f};
    float hpr[3] = { 117.0f, -24.0f, 0.0f};
    //float xyz[3] = {  0.11f, -15.65f, 13.41f};
    //float hpr[3] = {  90.0f,  -47.0f,  0.0f};

    dsSetViewpoint(xyz,hpr);
    dsSetSphereQuality(3);
}

/*** R[obN֐ ***/
static void nearCallback (void *data, dGeomID o1, dGeomID o2)
{
    int i,n;
    dBodyID b1 = dGeomGetBody(o1);
    dBodyID b2 = dGeomGetBody(o2);

    if (b1 && b2 && dAreConnectedExcluding(b1,b2,dJointTypeContact)) return;

    static const int N = 10;
    dContact contact[N];
    n = dCollide(o1,o2,N,&contact[0].geom,sizeof(dContact));
    if (n > 0)
    {
        for (i=0; i<n; i++)
        {
            if ((o1 == wheel[FRONT].geom) || (o2 == wheel[FRONT].geom) ||
                    (o1 == wheel[REAR].geom)  || (o2 == wheel[REAR].geom))
            {
                contact[i].surface.mode = dContactSoftERP | dContactSoftCFM | dContactApprox1;
                //contact[i].surface.mode = dContactSoftERP | dContactSoftCFM ;
                contact[i].surface.mu = 0.0; // dInfinity; // CW0
                contact[i].surface.soft_erp = 1.0;       // nʂERP
                contact[i].surface.soft_cfm = 1e-5;      // nʂCFM
            }
            else
            {
                // [Ur[̏Փ˔
                if (dGeomGetClass(o1) == dRayClass || dGeomGetClass(o2) == dRayClass)
                {
                    for (int k=0; k < BEAM_NUM; k++)
                    {
                        if (o1 == beam[k] || o2 == beam[k])
                        {
                            // ̂ȂȂ߂
                            if ((contact_pos[k][0] == -1) && (contact_pos[k][1] == -1)
                                    && (contact_pos[k][2]==-1))
                            {
                                for (int j=0; j < 3; j++)
                                    contact_pos[k][j] = contact[i].geom.pos[j];
                            }
                            else
                            {
                                return;
                            }
                        }
                    }
                    return; // ڐG_邾Ȃ̂Ŗ߂
                }

                contact[i].surface.mode = // dContactSlip1 | dContactSlip2 |
                    dContactSoftERP | dContactSoftCFM | dContactApprox1 | dContactMu2;
                contact[i].surface.mu  = 1.0; // dInfinity; // CW
                contact[i].surface.slip1    = 0.001;      // PC̊
                contact[i].surface.slip2    = 0.001;      // QC̊
                contact[i].surface.soft_erp = 1.0;       // nʂERP
                contact[i].surface.soft_cfm = 1e-5;      // nʂCFM
            }
            dJointID c = dJointCreateContact(world,contactgroup,&contact[i]); //ڐGWCg̐
            dJointAttach(c,b1,b2);                   // WCǧ
        }
    }
}

/*** L[{[hR}h̏֐ ***/
static void command(int cmd)
{
    switch (cmd)
    {
    case '1':
        float xyz1[3],hpr1[3];
        dsGetViewpoint(xyz1,hpr1);
        printf("xyz=%4.2f %4.2f %4.2f ",xyz1[0],xyz1[1],xyz1[2]);
        printf("hpr=%6.2f %6.2f %5.2f \n",hpr1[0],hpr1[1],hpr1[2]);
        while (1);
        break;
        //case 'j': v_r += 0.1; break;     // Eԗ,Oi
        //case 'f': v_l += 0.1; break;     // ԗ,Oi
        //case 'k': vcommand_r -= 0.1; break;     // Eԗ,i
        //case 'd': v_l -= 0.1; break;     // ԗ,i
    case 'b':
        v_l -= 0.1;
        v_r -= 0.1;
        break;     // Eԗ,Oi
    case 'f':
        v_l += 0.1;
        v_r += 0.1;
        break;     // ԗ,Oi
    case 'r':
        v_l += 0.1;
        v_r -= 0.1;
        break;     // Eԗ,i
    case 'l':
        v_l -= 0.1;
        v_r += 0.1;
        break;     // ԗ,i
    case 's':
        v_r = v_l = 0.0;
        break; //   ~
    case 'c':
        dBodySetPosition(ball.body,0,0,0.14);
        dBodySetLinearVel(ball.body,0,0,0);
        dBodySetAngularVel(ball.body,0,0,0);
        break;
    case 'u':
        LASER_ANGLE += 1 * M_PI/180;
        break;
    case 'd':
        LASER_ANGLE -= 1 * M_PI/180;
        break;
    case 'v':
        show = !show;
        break;
    default:
        printf("\nInput j,f,k,d,l,s,c,1:%d \n",cmd);
        break;
    }
}


// tB[hWn烍{bgWnւ̕ϊ  vO4.2
static void field2RobotPos(const dReal p[2], const dReal c[2], dReal theta,
                           dReal pr[2], dReal *r, dReal *phi)
{
    pr[0] =   (p[0]-c[0]) * cos(theta) + (p[1]-c[1]) * sin(theta);
    pr[1] = - (p[0]-c[0]) * sin(theta) + (p[1]-c[1]) * cos(theta);

    *r   = sqrt(pr[0] * pr[0] + pr[1] * pr[1]);  // ̂܂ł̋
    *phi = atan2(pr[0],pr[1]);                   // ̂̊px
}

static void drawRobot()
{
    // 
    dsSetColor(1.3,1.3,0.0);
    dsDrawCylinder(dGeomGetPosition(base.geom),
                   dGeomGetRotation(base.geom),BASE_L, BASE_R);

    // 
    dsSetColor(1.3, 1.3, 0.0);
    dsDrawCylinder(dBodyGetPosition(neck.body),
                   dBodyGetRotation(neck.body),NECK_L,NECK_R);

    // ZT
    dsSetColorAlpha(1.3, 1.3, 0.0, 0.8);
    dsDrawCylinder(dGeomGetPosition(sensor.geom),
                   dGeomGetRotation(sensor.geom),SENSOR_L, SENSOR_R);

    // [U[
    dsSetColor(0.0,0.0,0.0);
    dsDrawCylinder(dBodyGetPosition(laser.body),
                   dBodyGetRotation(laser.body),LASER_L, LASER_R);


    // ԗ
    dsSetColor(0.5,0.5,0.5);
    static dReal wheel_width = 0.02;
    for (int i=0; i< NUM; i++)
    {
        if ((i % 2) == 0)
        {
            if (i == FRONT) dsSetColor(1.3,1.3,1.3);
            else            dsSetColor(0.5,0.5,0.5);
            dsDrawCylinder(dBodyGetPosition(wheel[i].body),
                           dBodyGetRotation(wheel[i].body),wheel_width,0.95*WHEEL_SMALL_R);
        }
        else
        {
            dsSetColor(0.5,0.5,0.5);
            dsDrawCylinder(dBodyGetPosition(wheel[i].body),
                           dBodyGetRotation(wheel[i].body),wheel_width,WHEEL_BIG_R);
        }
    }
}

static void drawParticles()
{
    static int counter = 0;
    static dReal pos[1000][3];
    dMatrix3 R;
    dReal rb = 0.01;
    dsSetColor(1,0,0);
    dRSetIdentity(R);

    dReal r;
    dReal theta;

    for (int i=0; i < PARTICLE_NUM; i++)
    {
        if (counter == 0)
        {
            r = 2.0 * ((dReal) rand()/RAND_MAX -0.5);
            theta = M_PI * (dReal) rand()/RAND_MAX;
            pos[i][0] = r * cos(theta);
            pos[i][1] = r * sin(theta);
            pos[i][2] = 0.02;
        }
        dsDrawSphere(pos[i], R, rb);
    }
    counter++;
}

/*** [UXLisb`ɐU ***/
static void swingLaser(dReal target)
{
    static dReal kp = 20.0, fmax = 100;

    dReal tmp   = dJointGetHingeAngle(laser_joint);
    dReal diff  = target - tmp;
    if (diff >=   M_PI) diff -= 2.0 * M_PI; // diff2΂菬
    if (diff <= - M_PI) diff += 2.0 * M_PI; // diff-2΂傫
    dReal u     = kp * diff;

    dJointSetHingeParam(laser_joint, dParamVel,  u);
    dJointSetHingeParam(laser_joint, dParamFMax, fmax);
}


/*** ԗւ̐ ***/
static void controlWheel(dReal right, dReal left)
{
    double fMax = 100;  // őgN

    dJointSetHingeParam(joint[LEFT],  dParamVel, left);
    dJointSetHingeParam(joint[LEFT], dParamFMax, fMax);
    dJointSetHingeParam(joint[RIGHT],  dParamVel, right);
    dJointSetHingeParam(joint[RIGHT], dParamFMax, fMax);
}

/******* Oii֐go() **********/
static void go(double left_vel, double right_vel)
{
    v_l = left_vel; //[m/s]
    v_r = right_vel; //[m/s]
}
/******* Pɐ^iފ֐goStraight() **********/
static void goStraight(double vel)
{
    v_l = vel;
    v_r = vel;
}

/******* ]֐turn() **********/
static void turn(double ang_vel)
{
    v_l =  ang_vel;
    v_r = -ang_vel;
}

/******* Stop֐stop() ********/
static void stop()
{
    v_l = 0;
    v_r = 0;
}


/*** irQ[V lpO`vO@***/
// i20m,  90x], i20m, 90x]
void navigation(double vel)
{
    goStraight(vel);
}

/*** V~[V[v ***/
static void simLoop(int pause)
{
    double vel = 10.0;
    static double laser_angle = 0;
    static double diff = 1;

    emitLaser(); // [U[
    if (!pause)
    {
        controlWheel(v_r, v_l);

        if ((laser_angle >= M_PI/4) || (laser_angle <= - M_PI/4))
        {
            diff *= -1;
        }
        swingLaser(laser_angle += 5.0 * diff * M_PI/180.0);

        deadReckoning();
        //navigation(vel);

        dSpaceCollide(space,0,&nearCallback);
        dWorldQuickStep(world, STEP_SIZE);
        dJointGroupEmpty(contactgroup);
    }

    laser_receiveData(&urg, laser_data,laser_data_max);
    //for (int i=0; i < laser_data_max;i++) {
    //    printf("Laser dist[%d]=%d \n",i,laser_data[i]);
    //}

    //drawParticles();
    drawRobot();
    if (show) drawLaser();
    drawWall();
    steps++;
}

/*** `֐̐ݒ ***/
static void setDrawStuff()
{
    fn.version = DS_VERSION;
    fn.start   = &start;
    fn.step    = &simLoop;
    fn.command = &command;
    fn.path_to_textures = "../../drawstuff/textures";
}

/*** [UXLȉ ***/
void initLaser()
{
    laser_data = (long*)malloc(sizeof(long) * laser_data_max);
    if (laser_data == NULL)
    {
        perror("malloc");
        exit(1);
    }
    laser_setSkipLines(&urg,2);
    laser_requestData(&urg, URG_MD, URG_FIRST, URG_LAST);
}


/*** C֐ ***/
int main(int argc, char *argv[])
{
    setDrawStuff(); // `֐̐ݒ
    srand(time(NULL));
    dInitODE();
    world        = dWorldCreate();              // [h̐
    space        = dHashSpaceCreate(0);         // Xy[X̐
    contactgroup = dJointGroupCreate(0);        // ڐG_O[v̐
    ground       = dCreatePlane(space,0,0,1,0); // nʂ̐

    dWorldSetGravity(world, 0, 0, -9.8); // d͉x̐ݒ
    //dWorldSetCFM(world,1e-5); // CFM̐ݒ
    dWorldSetERP(world,1.0);  // ERP̐ݒ 0.8

    makeRobot(); // {bg̐
    makeWall(); // ǂ̐
    initLaser(); // [UXLȉ

    dsSimulationLoop(argc,argv,800,480,&fn); // V~[V[v

    dJointGroupDestroy(contactgroup);  // ڐG_O[v̔j
    dSpaceDestroy(space);              // Xy[X̔j
    dWorldDestroy(world);              // [h̔j
    dCloseODE();
    return 0;
}
